// Copyright (C) 2025  Jakob Bierwolf

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include <error.h>
#include <stdint.h>
#include <stdio.h>

#define N 2lu
const uint64_t MAX = (uint64_t)UINT32_MAX + 1;

typedef struct {
  uint32_t n[N];
} barr;

barr barr_new(uint64_t value) {
  barr number = {.n = {0}};
  for (int i = 0; value; i++) {
    if (i >= N)
      error(1, 0,
            "barr_new(): Number overflow, can only store %lu * "
            "sizeof(uint32_t) bytes",
            N);

    number.n[i] += value % MAX;
    value /= MAX;
  }
  return number;
}

void barr_print(barr a) {
  int omit_leading_zeros = 0;
  // int omit_leading_zeros = 1;
  printf("0x");
  for (int i = N - 1; i >= 0; i--) {
    if (!omit_leading_zeros || a.n[i] || !i) {
      printf("%08x", a.n[i]);
      omit_leading_zeros = 0;
    }
  }
}

barr barr_add(barr a, barr b) {
  barr sum = barr_new(0);
  for (int i = 0; i < N; i++) {
    uint64_t s = (uint64_t)sum.n[i] + (uint64_t)a.n[i] + (uint64_t)b.n[i];
    if (s < MAX)
      sum.n[i] = s;
    else if (i + 1 < N) {
      sum.n[i] = s % MAX;
      sum.n[i + 1] += s / MAX;
    } else {
      error(1, 0,
            "barr_add(): Number overflow, can only store %lu * "
            "sizeof(uint32_t) bytes",
            N);
    }
  }
  return sum;
}

barr barr_sub(barr a, barr b) {
  barr diff = barr_new(0);
  for (int i = 0; i < N; i++) {
    int32_t d = (int32_t)diff.n[i] + (int32_t)a.n[i] - (int32_t)b.n[i];
    diff.n[i] = d;
    if (d < 0) {
      if (i + 1 < N) {
        diff.n[i] = d + MAX;
        diff.n[i + 1] += d / MAX;
      } else {
        error(1, 0,
              "barr_sub(): Number overflow, can only store %lu * "
              "sizeof(uint32_t) bytes",
              N);
      }
    }
  }
  return diff;
}

int main(void) {
  // barr a = barr_new(0xffffffffffffffff);
  // barr b = barr_new(0x0000000000000001);
  barr a = barr_new(0x00000000ffffffff);
  barr b = barr_new(0x0000000000000001);
  barr c = barr_add(a, b);
  printf("  ");
  barr_print(a);
  printf("\n+ ");
  barr_print(b);
  printf("\n--------------------\n= ");
  barr_print(c);
  printf("\n");
}
